/*******************************************************************************
 * $Id: memsApp.cpp,v 1.48 2004/02/28 00:01:31 darling Exp $
 * This file is copyright 2002-2004 Aaron Darling.  All rights reserved.
 * Please see the file called COPYING for licensing, copying, and modification
 * rights.  Redistribution of this file, in whole or in part is prohibited
 * without express permission.
 ******************************************************************************/

#include <sstream>
#include <stdexcept>
#include "libMems/Matrix.h"
#include "libMems/NumericMatrix.h"
#include "getopt.h"
#include "gn/gnSequence.h"
#include "libMems/DNAFileSML.h"
#include "libMems/MemHash.h"
#include "libMems/MaskedMemHash.h"
#include "libMems/Aligner.h"
#include "libMems/MatchList.h"
#include "libMems/MemSubsets.h"
#include "libMems/RepeatHash.h"
#include "libMems/Interval.h"
#include "libMems/IntervalList.h"
#include "libMems/gnAlignedSequences.h"
#include "libMems/Islands.h"


void print_usage( const char* pname ){
	cerr << "Usage:" << endl;
	cerr << pname << " [-lr] [-m mer size] [-d density] [-s size] "
		<< "[-i input file] [-t stats file] [-o output file] <seq1 filename> <sml1 filename> ... "
		<< " <seqN filename> <smlN filename>" << endl;
	cerr << "Options:" << endl;
	cerr << "\t    --mums Find MUMs only, do not attempt to determine locally collinear blocks (LCBs)\n";
	cerr << "\t-r, --no-recursion Don't perform a recursive alignment" << endl;
	cerr << "\t    --no-lcb-extension If determining LCBs, don't attempt to extend the LCBs\n";
	cerr << "\t-m, --seed-size=<number> Initial seed match size, default is log_2( average seq. length )" << endl;
	cerr << "\t    --eliminate-inclusions Eliminate linked inclusions in subset matches.\n";
	cerr << "\t-s, --weight=<number> Minimum LCB weight in base pairs per sequence" << endl;
	cerr << "\t-i, --match-input=<file> Use specified match file instead of searching for matches\n";
	cerr << "\t    --lcb-input=<file> Use specified lcb file instead of constructing LCBs\n";
	cerr << "\t-o, --output=<file> Output file name.  Prints to screen by default" << endl;
	cerr << "\t    --id-matrix=<file> Generate LCB stats and write them to the specified file\n";
	cerr << "\t    --island-size=<number> Find islands larger than the given number\n";
	cerr << "\t    --island-output=<file> Output islands the given file (requires --island-size)\n";
	cerr << "\t    --backbone-size=<number> Find stretches of backbone longer than the given number of b.p.\n";
	cerr << "\t    --max-backbone-gap=<number> Allow backbone to be interrupted by gaps up to this length in b.p.\n";
	cerr << "\t    --backbone-output=<file> Output islands the given file (requires --island-size)\n";
	cerr << "\t    --coverage-output=<file> Output a coverage list to the specified file (- for stdout)\n";
	cerr << "\t-R, --repeats Generates a repeat map.  Only one sequence can be specified\n";
	cerr << "\t    --max-clustal-length=<number> Maximum number of base pairs to attempt aligning with ClustalW (Default is 9000)\n";
	cerr << "\t    --min-recursive-gap-length=<number> Minimum size of gaps that Mauve will perform recursive MUM anchoring on (Default is 200)\n";
	cerr << "\t    --clustal-output-dir=<directory> Outputs a clustal format gapped alignment for each LCB to a given directory\n";
	cerr << "\t    --output-guide-tree=<file> Write out the guide tree used for CLUSTALW alignment to the designated file\n";
	cerr << "\t    --output-alignment=<file> Write out a gapped alignment to the designated file\n";
	cerr << endl;
}

class MLDeleter {
public:
	MLDeleter( MatchList& ml ) : mlist( ml ) {}
	~MLDeleter(){ mlist.Clear(); }
private:
	MatchList& mlist;
};

int viceroy( int argc, char* argv[] );
#define NELEMS(a) ( sizeof( a ) / sizeof( *a ) )
#if defined(__MWERKS__) && defined(__GNDEBUG__)
int main( int argc, char* argv[] ){
#else
int viceroy( int argc, char* argv[] ){
#endif

	char* m_argv[] = {
		"mauveAligner",
		"--lcb-input=isle.out",
		"--backbone-size=100",
		"--max-backbone-gap=100",
		"--backbone-output=isle_bb.out",
		"diarrhea\\EDL933.fas",
		"diarrhea\\EDL933.fas.sml",
		"diarrhea\\ecolim52.fas",
		"diarrhea\\ecolim52.fas.sml",
		"diarrhea\\o157sakai.fas",
		"diarrhea\\o157sakai.fas.sml",
	};

	int m_argc = NELEMS( m_argv );
	int rval = 0;
#if defined(__MWERKS__) && defined(__GNDEBUG__)
	rval = viceroy( m_argc, m_argv );
#endif
	return rval;
}

/**
 * This basic application demonstrates how to use libMems to produce full scale multiple
 * genomic alignments.  First the command line is parsed to get the names of data files
 * and user specified options.  Next each sequence and its corresponding sorted mer list
 * are loaded.  If the sorted mer list fails to load a new one is created.  
 * If it is necessary to find matches in the sequences instead of loading them, each 
 * sequence and SML are added to a MemHash which searches for exact matches.  
 * Then LCBs are found if the user requested it.  Finally, either the MatchList or the
 * LCB list is written to disk.
 */
#if defined(__MWERKS__) && defined(__GNDEBUG__)
int viceroy( int argc, char* argv[] ){
#else
int main( int argc, char* argv[] ){
#endif
try{
	if( argc <= 0 ){
		print_usage( "mauveAligner" );
		return -1;
	}
	if( argc == 1 ){
		print_usage( argv[0] );
		return -1;
	}

	//
	// definitions of the variables that can be set by the user on the command line:
	//
	vector<string> seq_files;
	vector<string> sml_files;
	vector<gnSequence*> seq_table;
	vector<DNAFileSML*> sml_table;
	uint mer_size = 0;	// Use default settings
	boolean recursive = true;
	boolean lcb_extension = true;
	boolean create_LCBs = true;
	boolean calculate_coverage = false;
	int64 LCB_size = -1;
	string output_file = "";
	boolean read_matches = false;
	boolean read_lcbs = false;
	boolean find_repeats = false;
	boolean print_stats = false;
	boolean eliminate_inclusions = false;
	boolean eliminate_overlaps = false;
	boolean nway_filter = false;
	string match_input_file = "";
	string lcb_stats_file = "";
	string island_file = "";
	string lcb_file = "";
	string tree_filename = "";
	string coverage_list_file = "";
	boolean output_alignment = false;
	string clustal_output_dir = "";
	string alignment_output_file = "";
	string match_log = "";
	string offset_log = "";
	string merge_log = "";
	// island related
	uint island_size = 0;
	uint island_break_min = 0;
	// backbone related
	uint backbone_size = 0;
	uint max_backbone_gap = 0;
	int64 min_r_gap_length = -1;
	string backbone_file = "";
	boolean output_backbone = false;

	uint seqI;
	boolean print_version = false;
	int max_clustal_length;
	// parse command line with gnu getopt
	int opt;
	int ac = argc;
	char** av = argv;
	// 'm' mer size
	// 'r' recursive
	const char* short_args= "m:s:o:i:t:I:F:M:L:O:vrRE";
	struct option long_opts[] = {
		{"mums", no_argument, NULL, 'l'},
		{"no-recursion", no_argument, NULL, 'r'},
		{"no-lcb-extension", no_argument, NULL, 'e'},
		{"seed-size", required_argument, NULL, 'm'},
		{"weight", required_argument, NULL, 's'},
		{"output", required_argument, NULL, 'o'},
		{"eliminate-inclusions", no_argument, NULL, 'E' },
		{"eliminate-overlaps", no_argument, NULL, 'V' },
		{"n-way-filter", no_argument, NULL, 'N' },
		{"match-input", required_argument, NULL, 'i'},
		{"lcb-input", required_argument, NULL, 'L'},
		{"output-alignment", optional_argument, NULL, 'O'},
		{"id-matrix", optional_argument, NULL, 't'},
		{"island-size", required_argument, NULL, 'I'},
		{"island-output", required_argument, NULL, 'F'},
		{"island-break-min", required_argument, NULL, 'M'},
		{"backbone-size", required_argument, NULL, 'B'},
		{"max-backbone-gap", required_argument, NULL, 'g'},
		{"backbone-output", optional_argument, NULL, 'b'},
		{"coverage-output", optional_argument, NULL, 'c'},
		{"repeats", no_argument, NULL, 'R'},
		{"max-clustal-length", required_argument, NULL, 'C'},
		{"min-recursive-gap-length", required_argument, NULL, 'P'},
		{"output-guide-tree", required_argument, NULL, 'T'},
		{"clustal-output-dir", required_argument, NULL, 'D'},
		{"match-log", required_argument, NULL, 'A'},
		{"offset-log", required_argument, NULL, 'S'},
		{"merge-match-log", required_argument, NULL, 'G'},
		{"version", no_argument, NULL, 'v'},
		{0, 0, 0, 0}	// for correct termination of option list
						// getopt_long can segfault without this
	};

	int indexptr;
	while( (opt = getopt_long( ac, av, short_args, long_opts, &indexptr )) != EOF ){
		switch( opt ){
			case 'l':
				create_LCBs = false;
				break;
			case 'r':
				recursive = false;
				break;
			case 'e':
				lcb_extension = false;
				break;
			case 'm':
				mer_size = atoi( optarg );
				break;
			case 's':
				LCB_size = atol( optarg );
				break;
			case 'o':
				output_file = optarg;
				break;
			case 'i':
				read_matches = true;
				match_input_file = optarg;
				break;
			case 't':
				print_stats = true;
				if( optarg != NULL )
					lcb_stats_file = optarg;
				break;
			case 'I':
				island_size = atoi( optarg );
				break;
			case 'F':
				island_file = optarg;
				break;
			case 'L':
				lcb_file = optarg;
				read_lcbs = true;
				break;
			case 'M':
				island_break_min = atoi( optarg );
				break;
			case 'c':
				if( optarg != NULL )
					coverage_list_file = optarg;
				calculate_coverage = true;
				break;
			case 'R':
				find_repeats = true;
				break;
			case 'E':
				eliminate_inclusions = true;
				break;
			case 'O':
				output_alignment = true;
				if( optarg != NULL )
					alignment_output_file = optarg;
				break;
			case 'C':
				max_clustal_length = atoi( optarg );
				ClustalInterface::getClustalInterface().SetMaxClustalLength( max_clustal_length );
				break;
			case 'T':
				tree_filename = optarg;
				break;
			case 'D':
				clustal_output_dir = optarg;
				break;
			case 'A':
				match_log = optarg;
				break;
			case 'S':
				offset_log = optarg;
				break;
			case 'G':
				merge_log = optarg;
				break;
			case 'g':
				max_backbone_gap = atoi( optarg );
				break;
			case 'B':
				backbone_size = atoi( optarg );
				break;
			case 'b':
				backbone_file = optarg;
				output_backbone = true;
				break;
			case 'v':
				print_version = true;
				break;
			case 'N':
				nway_filter = true;
				break;
			case 'V':
				eliminate_overlaps = true;
				break;
			case 'P':
				min_r_gap_length = atol( optarg );
				break;
			default:
				print_usage( argv[0] );
				return -1;
		}
	}
	// now read in the seq and sml file names from av
	boolean seq_name_arg = true;
	for( int optI = optind; optI < argc; optI++ ){
		if( seq_name_arg )
			seq_files.push_back( av[ optI ] );
		else
			sml_files.push_back( av[ optI ] );
		seq_name_arg = !seq_name_arg;
	}
	
	// print the version if the user requested it
	if( print_version ){
		cerr << "mauveAligner " << " build date " << __DATE__ << " at " << __TIME__ << endl;
	}
	
	if( ( island_size != 0 && island_file == "" ) || ( island_size == 0 && island_file != "" ) ){
		cerr << "Error: Both --island-file and --island-size must be specified\n";
		return -1;
	}
	if( find_repeats ){
		if( create_LCBs || read_matches || read_lcbs || calculate_coverage || 
		    island_file != "" || island_size != 0 || recursive || lcb_stats_file != "" ){
			cerr << "A paramater has been specified that is incompatible with repeat list generation\n";
			return -1;
		}
	}
	if( seq_files.size() > MAX_SEQ_COUNT ){
		cerr << "This copy of mauveAligner was compiled to align a maximum of " << MAX_SEQ_COUNT << " sequences\n";
		return -1;
	}
	// done parsing and checking command line options
	// Start doing the work
	
	MatchList match_list;
	MLDeleter deleter( match_list );
	
	if( seq_files.size() == 1 && sml_files.size() == 0 ){
		match_list.LoadMFASequences( seq_files[0], mer_size, &cout, find_repeats || ( !read_lcbs && !read_matches ));
	}else if( seq_files.size() != sml_files.size() ){
		cerr << "Error: Each sequence file must have a corresponding SML file specified.\n";
		return -1;
	}else{
		match_list.seq_filename = seq_files;
		match_list.sml_filename = sml_files;
		match_list.LoadSequences( mer_size, &cout, find_repeats || !read_matches || ( !read_lcbs && !read_matches ));
	}

	ostream* match_out;
	if( output_file != "" ){
		ofstream* match_out_file = new ofstream( output_file.c_str() );
		if( !match_out_file->is_open() ){
			cerr << "Error opening " << output_file << endl;
			return -2;
		}
		match_out = match_out_file;
	}else
		match_out = &cout;
	
	// search for repetitive regions
	if( find_repeats ){
		RepeatHash repeat_finder;
		repeat_finder.LogProgress( &cout );
		repeat_finder.FindMatches( match_list );
		match_list.WriteList( *match_out );
		match_out->flush();
		return 0;
	}
	
	// read matches if the user requested it
	if( read_matches ){
		ifstream match_in( match_input_file.c_str() );
		if( !match_in.is_open() ){
			cerr << "Error opening " << match_input_file << endl;
			return -2;
		}
		try{
			match_list.ReadList( match_in );
		}catch( gnException& gne ){
			cerr << "Error reading " << match_input_file << "\nPossibly corrupt file or invalid file format\n";
			return -2;
		}
		if( seq_files.size() > 1 )
			match_list.seq_filename = seq_files;
		else if( match_list.seq_table.size() == 0 )
			// fill seq_table with empty sequences
			for( seqI = 0; seqI < match_list.seq_filename.size(); seqI++ )
				match_list.seq_table.push_back( new gnSequence() );
	}else if ( !read_lcbs ){
		// get full subset matches
		MemHash match_finder;
		match_finder.LogProgress( &cout );
		fstream match_log_out;
		if( match_log != "" ){
			match_log_out.open( match_log.c_str(), ios::in | ios::out );
			if( !match_log_out.is_open() ){
				cerr << "Error opening " << match_log << endl;
				return -1;
			}
			match_finder.SetMatchLog( &match_log_out );
			// append to whatever's already in the file
			match_log_out.seekg( 0, ios::end );
		}
		fstream offset_log_out;
		vector< gnSeqI > offset_start;
		for( seqI = 0; seqI < match_list.seq_table.size(); seqI++ )
			offset_start.push_back( 0 );
		
		if( offset_log != "" ){
			offset_log_out.open( offset_log.c_str(), ios::in | ios::out );
			if( !offset_log_out.is_open() ){
				cerr << "Error opening " << offset_log << endl;
				return -1;
			}
			match_finder.SetOffsetLog( &offset_log_out );
			string last_line;
			string cur_line;
			while( getline( offset_log_out, cur_line ) ){
				last_line = cur_line;
			}
			if( last_line != "" ){
				stringstream cur_off_stream( last_line );
				for( seqI = 0; seqI < match_list.seq_table.size(); seqI++ )
					cur_off_stream >> offset_start[ seqI ];
			}
			offset_log_out.clear();
		}
		ifstream merge_log_in;
		if( merge_log != "" ){
			merge_log_in.open( merge_log.c_str() );
			if( !merge_log_in.is_open() ){
				cerr << "Error opening " << merge_log << endl;
				return -1;
			}

			for( seqI = 0; seqI < match_list.seq_table.size(); seqI++ ){
				if( !match_finder.AddSequence( match_list.sml_table[ seqI ], match_list.seq_table[ seqI ] ) ){
					ErrorMsg( "Error adding " + match_list.seq_filename[seqI] + "\n");
					return -1;
				}
			}
			match_finder.LoadFile( merge_log_in );
			match_finder.GetMatchList( match_list );
		}else
			match_finder.FindMatchesFromPosition( match_list, offset_start );
		match_log_out.close();
		offset_log_out.close();
	}
		

	// write out a match list if the user doesn't want LCBs
	if( !create_LCBs && !read_lcbs){
		if( eliminate_inclusions ){
			// Remove linked inclusions
			SubsetInclusionRemover sir;
			sir.EliminateLinkedInclusions( match_list );
		}
		if( eliminate_overlaps ){
			Aligner::EliminateOverlaps( match_list );
		}

		if( nway_filter ){
			match_list.MultiplicityFilter( match_list.seq_table.size() );
		}

		match_list.WriteList( *match_out );
		match_out->flush();
		
		// output a guide tree or a coverage list if necessary
		// beware that selecting the nway filter above will cause the guide tree
		// and coverage lists to be incorrect
		vector< pair< uint64, uint64 > > coverage_list;
		if( tree_filename != "" || calculate_coverage ){
			if( !eliminate_inclusions ){
				// Remove linked inclusions if it hasn't already been done
				SubsetInclusionRemover sir;
				sir.EliminateLinkedInclusions( match_list );
			}
			// only count each base pair once!
			if( !eliminate_overlaps )
				Aligner::EliminateOverlaps( match_list );

			allocateDetailList( match_list, coverage_list );
			getLCBDetailList( match_list, coverage_list );
		}
		if( calculate_coverage ){
			if( coverage_list_file == "" || coverage_list_file == "-" )
				PrintDetailList( match_list.seq_table.size(), coverage_list, cout );
			else{
				ofstream coverage_file( coverage_list_file.c_str() );
				if( !coverage_file.is_open() ){
					cerr << "Error opening " << coverage_list_file << endl;
					return -2;
				}
				PrintDetailList( match_list.seq_table.size(), coverage_list, coverage_file );
			}
		}

		if( tree_filename != "" ){
			NumericMatrix< double > distance;
			DistanceMatrix( match_list.seq_table.size(), coverage_list, distance );
			ClustalInterface& ci = ClustalInterface::getClustalInterface();
			if( tree_filename == "" ){
				wxString temp_treename = wxGetTempFileName("guide_tree");
				tree_filename = temp_treename.c_str();
			}
			ci.SetDistanceMatrix( distance, tree_filename );
		}

		return 0;
	}
	
	// at this point any SortedMerLists used to identify the initial set of MUMs
	// are no longer necessary.  Free them
	for( uint smlI = 0; smlI < match_list.sml_table.size(); smlI++ ){
		delete match_list.sml_table[ smlI ];
	}
	match_list.sml_table.clear();
	
	// Align the sequences if necessary
	if( LCB_size < 0 ){
		// calculate a default LCB weight, 3 times the mer size times the seq. count
		if( mer_size <= 0 )	
			mer_size = MatchList::GetDefaultMerSize( match_list.seq_table );
		LCB_size = mer_size * 3 * match_list.seq_table.size();
	}else{
		// adjust the LCB weight for the number of sequences being aligned
		LCB_size *= match_list.seq_table.size();
	}
	// check that LCB_size can be set appropriately
	if( create_LCBs && LCB_size < 0 ) {
		
		cerr << "A minimum LCB size greater than 0 must be specified in order to create LCBs.\n";
		return -1;
	}

	Aligner aligner( match_list.seq_table.size() );

	if( min_r_gap_length >= 0 ){
		aligner.SetMinRecursionGapLength( min_r_gap_length );
	}
	
	IntervalList interval_list;
	interval_list.seq_table = match_list.seq_table;
	interval_list.seq_filename = match_list.seq_filename;
	if( lcb_file == "" ){

		try{
			aligner.align( match_list, interval_list, 0, LCB_size, recursive, lcb_extension, tree_filename );
		}catch( gnException& gne ){
			cerr << gne << endl;
		}
		interval_list.WriteList( *match_out );
		match_out->flush();

	}else if( read_lcbs ){
		ifstream lcb_input( lcb_file.c_str() );
		if( !lcb_input.is_open() ){
			cerr << "Error opening " << lcb_file << endl;
			return -2;
		}
		try{

			interval_list.seq_table = match_list.seq_table;
			interval_list.seq_filename = match_list.seq_filename;
			interval_list.ReadList( lcb_input );

		}catch( gnException& gne ){
			cerr << gne << endl;
			cerr << "Error reading " << lcb_file << "\\nPossibly corrupt file or invalid file format\n";
			return -2;
		}
	}

	if( output_alignment ){
		if( alignment_output_file == "" || alignment_output_file == "-" ){
			interval_list.WriteStandardAlignment( cout );
		}else{
			ofstream align_out( alignment_output_file.c_str() );
			if( !align_out.is_open() ){
				cerr << "Error opening " << alignment_output_file << endl;
				return -1;
			}
			interval_list.WriteStandardAlignment( align_out );
			align_out.close();
		}
	}
	uint lcbI;
	
	// output alignments in clustalW format if the user asked for it
	if( clustal_output_dir != "" ){
		wxString cwd = wxGetCwd();
		wxString wxclustal_output_dir( clustal_output_dir.c_str() );
		if( !wxDirExists( wxclustal_output_dir ) ){
			wxMkdir( wxclustal_output_dir, 755 );
		}
		wxSetWorkingDirectory( wxclustal_output_dir );
		for( lcbI = 0; lcbI < interval_list.size(); lcbI++ ){
			gnAlignedSequences gnas;
			interval_list[ lcbI ].GetAlignedSequences( gnas, match_list.seq_table );
			ostringstream oss;
			oss << "lcb_" << lcbI << ".txt";
			ofstream clustal_lcb_out( oss.str().c_str(), ios::trunc );
			if( !clustal_lcb_out.is_open() ){
				cerr << "Error opening " << oss.str() << endl;
				return -1;
			}
			gnas.outputClustalW( clustal_lcb_out );
		}
		wxSetWorkingDirectory( cwd );
	}

	vector< pair< uint64, uint64 > > detail_list;
	allocateDetailList( interval_list, detail_list );
	for( uint lcbI = 0; lcbI < interval_list.size(); lcbI++ )
		getLCBDetailList( interval_list[ lcbI ], detail_list );
	PrintDetailList( interval_list.seq_table.size(), detail_list, cout );
	
	//
	// output an identity matrix if requested
	//
	if( print_stats ){
		ostream* stats_out;
		if( lcb_stats_file == "" || lcb_stats_file == "-" ){
			stats_out = &cout;
		}else{
			ofstream* stats_out_file = new ofstream( lcb_stats_file.c_str() );
			if( !stats_out_file->is_open() ){
				cerr << "Error opening " << lcb_stats_file << endl;
				return -1;
			}
			stats_out = stats_out_file;
		}
		NumericMatrix< double > identity;
		IdentityMatrix( interval_list, identity );
		identity.print( *stats_out );
		if( lcb_stats_file == "" || lcb_stats_file == "-" ){
			delete stats_out;
		}
	}

	//
	// output backbone if it was requested
	//
	if( output_backbone ){
		ostream* backbone_out;
		if( backbone_file != "" ){
			ofstream* backbone_out_file = new ofstream( backbone_file.c_str() );
			if( !backbone_out_file->is_open() ){
				cerr << "Error opening " << backbone_file << endl;
				return -1;
			}
			backbone_out = backbone_out_file;
		}else
			backbone_out = &cout;

		vector< ClustalResult > backbone_data;
		simpleFindBackbone( interval_list, backbone_size, max_backbone_gap, backbone_data );
		outputBackbone( backbone_data, *backbone_out );
		if( backbone_file != "" ){
			delete backbone_out;
		}
	}

	//
	// output islands if they were requested
	//
	if( island_file != "" ){
		ostream* island_out;
		if( island_file == "-" )
			island_out = &cout;
		else{
			ofstream* island_out_file = new ofstream( island_file.c_str() );
			if( !island_out_file->is_open() ){
				cerr << "Error opening " << island_file << endl;
				return -1;
			}
			island_out = island_out_file;
		}
		simpleFindIslands( interval_list, island_size, *island_out );
		findIslandsBetweenLCBs( interval_list, island_size, *island_out );

		if( island_file != "-" ){
			delete island_out;
		}
	}

}catch( gnException& gne ) {
	cerr << "Unhandled gnException: " << gne << endl;
}catch( exception& e ) {
	cerr << "Unhandled exception: " << e.what() << endl;
}catch(...){
	cerr << "Unknown exception occurred.\n";
}

	return 0;
}
